#!/usr/bin/env python3
#
# Create .dat and .csv files from the summaries of FIG experimentation
#
# Author:  Carlos E. Budde
# Date:    21.10.2021
# License: GPLv3
#

import os
import re
from typing import List,TextIO
from io import TextIOWrapper
from sys import argv, exit
from mmap import mmap, ACCESS_READ
from numpy import inf, abs, mean, median, std, array


if len(argv) != 5:
	print("[ERROR] Must call with 1. summary filename, 2. suffix pattern, 3. csv filename, 4. comma-separated list with all run types including ifuns")
	exit(1)


########################################
##
##  CONFIGURATION REGEXes
##
# Whatever is considered a number: 1, 2.34, 1.23e-4...
ARBITRARY_NUM = "inf|[0-9]+(?:\.[0-9]+(?:(?:e|E)(?:\-|\+)[0-9]*)*)*"
# Primary header: a run type, e.g. "FLAT", "RES2_ES", "RES_8"...
HEAD1_REGEX = br"\n[A-Z]+[0-9]*(?:_[A-Z]*[0-9]*)*\n"
# Secondary header: an importance function, e.g. "IFUN: STR", "IFUN: MCS"...
HEAD2_REGEX = bytes("IFUN: .*","utf8")


########################################
##
##  SCRIPT
##

def trimb(bstr: bytes) -> str:
	"""Return string contained in byte-array"""
	return str(bstr)[2:-3]

def eat_spaces(fc: mmap) -> mmap:
	"""Advance fc from its current position to the first non-space line"""
	pos = fc.tell()
	line = fc.readline()
	while line.isspace():
		pos = fc.tell()
		line = fc.readline()
	fc.seek(pos)
	return fc

def find_numbers(fc: mmap) -> List[str]:
	"""
	Return all numbers found from the current position of fc,
	using empty lines as delimiters:
	0. skip initial empty lines,
	1. eat all that follows, until the next empty line,
	2. stop and leave fc at that empty-line position,
	3. return the numbers from what was eaten (see ARBITRARY_NUM).
	"""
	numbers = ""
	fc = eat_spaces(fc)
	line = fc.readline()
	while not line.isspace():
		numbers += " " + trimb(line)
		line = fc.readline()
	return re.findall(ARBITRARY_NUM, numbers)

def robust_stats(values: List[float]) -> (float,float,int):
	"""
	Filter values passed, to keep those strictly between 0.0 and 1.0,
	and then also discard those which are too far away from the median.
	From the resulting "robust subset" return the mean, stdev, and size.
	Robust mean: https://stackoverflow.com/a/16562028
	"""
	data = array([ float(v) for v in values if 0.0 < float(v) < 1.0 ])
	num_valid_datapoints = len(data)
	if num_valid_datapoints < 1:
		return 0,inf,0
	dev = abs(data - median(data))
	mdev = median(dev)
	s = dev/mdev if mdev != 0.0 else 0.0
	m = 2.0  # outlier level: the greater, the fewer values filtered
	filtered_values = [data[s<m]]
	if len(filtered_values) > 0:
		robust_mean = mean(filtered_values)
		robust_stdev = std(filtered_values)
	else:
		robust_mean = 0
		robust_stdev = inf
	return robust_mean,robust_stdev,num_valid_datapoints

def extract_mean(fc: mmap, cnt: str, fun: str, df: TextIOWrapper) -> float:
	stats = [ "{:.2e}".format(v) for v in robust_stats(find_numbers(fc)) ]
	stats[-1] = str(int(float(stats[-1])))
	df.write(" ".join([cnt,fun] + stats))
	df.write("\n")
	return float(stats[0])

def read_experiment(fc: mmap, run: str, df: TextIOWrapper) -> List[tuple]:
	"""
	Read a whole experiment run, e.g. FE_8 with all importance functions.
	Write (precision) data read in df, formatted for gnuplot CI plotting.
	Also write in csv file the precision achieved for this model instance.
	NOTE: in the tuples, the first value has the pattern "algo_thresh_ifun",
	e.g. "FE_8_STR" where FE is algo, 8 is thresh, and STR is ifun.
	"""
	cnt = 0
	pmeans = []
	# Find run line in fc file
	fc.seek(0)
	fc.seek(fc.find(bytes(run,"utf8")))
	line = fc.readline()
	df.write("# fun avg      stdev    N\n")
	# His Lordship crude Monte Carlo is special
	if run == "CMC":
		# Dump in df the mean,stdev,count of a robust subset of the values
		return [ tuple(["CMC",extract_mean(fc, str(cnt), "CMC", df)]) ]
	# In all other cases, for each IFUN ...
	fc = eat_spaces(fc)
	line = fc.readline()
	while type(None) != type(re.match(HEAD2_REGEX,line)):
		cnt += 1  # leave this here
		# ... identify the IFUN name/acronym ...
		IFUN = re.findall(".*:\s*([A-Z]+)", trimb(line))[0]
		# ... dump in df the mean,stdev,count of the (robust) values ...
		pmeans += [ tuple([run+"_"+IFUN, extract_mean(fc, str(cnt), IFUN, df)]) ]
		# ... and move on to the next IFUN
		fc = eat_spaces(fc)
		line = fc.readline()
	return pmeans

def homogenise_array(all_pmeans: List[List[tuple]]) -> int:
	"""
	Some runs may have failed, truncating the list of results returned.
	This function makes all lists in the array have the same number of
	elements (equal to the max) to avoid processing issues.
	"""
	# His Lordship crude Monte Carlo will go fishing
	all_pmeans.pop(RUNS.index("CMC"))
	# For the rest, it's time to work
	longest = max(len(pmeans) for pmeans in all_pmeans)
	homogeneous_pmeans = []
	for pmeans in all_pmeans:
		if len(pmeans) == 0:
			all_pmeans.remove(pmeans)  # no values, just delete
		elif len(pmeans) < longest:
			IFUN,_ = pmeans[0]
			pmeans += [ tuple([IFUN,0]) ] * (longest-len(pmeans))  # few values: pad
		# Remove "algo_thresh_" from the first element of the tuples,
		# to leave only the "ifun" (see read_experiment())
		pmeans_new = []
		for tup in pmeans:
			IFUN = tup[0].split("_")[-1]
			pmeans_new += [tuple([IFUN,tup[1]])]
		homogeneous_pmeans += [pmeans_new]
	return longest,homogeneous_pmeans

# Process this summary file
FNAME = argv[1]
SUFFIX = argv[2]
CSV = argv[3]
ORDERED_RUNS = sorted(set(argv[4].split(",")))
print("Processing " + FNAME)
#
with open(CSV,"a") as csv, open(FNAME) as f, mmap(f.fileno(), 0, access=ACCESS_READ) as fc:
	# Identify experiments ran
	RUNS = [ str(r)[4:-3] for r in set(re.findall(HEAD1_REGEX, fc)) ]
	RUNS.sort()
	PREFIX = FNAME[ : FNAME.find(SUFFIX) ]
	all_pmeans = []
	# Build a .dat file for each experiment
	for run in RUNS:
		DATA_FNAME = PREFIX + "_" + run + ".dat"
		dfile = open(DATA_FNAME, "w")
		print(" - " + run + " results", end=":  \t")
		all_pmeans += [ read_experiment(fc, run, dfile) ]
		dfile.close()
		print(DATA_FNAME)
	fc.close()
	# We have all the info to create the CSV. First the header (if needed)
	if os.stat(CSV).st_size==0:
		csv.write("model_prop,instance," + ",".join(ORDERED_RUNS) + "\n")
	# Now, for this model instance...
	# (assumes file name in format MODEL_INSTANCE_PROPERTY, e.g. "RC_3_rel")
	parts = PREFIX.split("_")
	csv.write(parts[0]+"_"+parts[2] +","+ parts[1] +",")
	# ...dump the (mean) precisions computed for each run, in order
	VALS = dict([ tuples for pmeans in all_pmeans for tuples in pmeans ])
	PRECISIONS = [ str(VALS[r]) if VALS.get(r) is not None else "" for r in ORDERED_RUNS ]
	csv.write(",".join(PRECISIONS) + "\n")
	# For the grand finale, show the precisions of the rare event runs
	prec_ref_vals = []
	NUM_IFUNS,all_pmeans = homogenise_array(all_pmeans)
	for i in range(NUM_IFUNS):
		IFUN = all_pmeans[0][i][0]
		pmeans_i = [ tup[1] for tup in [ l[i] for l in all_pmeans ] ]
		prec_ref_vals += [ [IFUN] + list(robust_stats(pmeans_i))[0:2] ]
	print(" Created "+str(len(RUNS))+" data files with "
		 + str(len(prec_ref_vals)) +" functions each (except CMC).\n"
		 + " The mean,stdev precision of the CIs achieved with each IFUN is:")
	for ifun,m,s in prec_ref_vals:
		print("   * " + ifun + ": {:1.2e}".format(m) + "," + "{:1.2e}".format(s))
	
