#!/usr/bin/env bash
#
# Bash script to summarise and plot results from FIG experimentation
#
# Author:  Carlos E. Budde
# Date:    21.10.2021
# License: GPLv3
#
#   NOTE: this script should be sourced from another script
#   Cool site: http://wiki.bash-hackers.org

SUFFIX="summary.txt";  # make extension different than raw FIG output files
LOG="plotting.log";    # IDEM above
PLVL="95%";            # confidence level for which precision will be plotted
FAIL_VALUE="0.05";      # value of "extra x ticks" in scatterplots.tex

CWD="`pwd`";
cd "$DIR_RESULTS";

# Configure script based on the running mode (FULL vs. DEMO)
if [ -z ${DEMO_RUN+x} ] || [ $DEMO_RUN -ne 1 ];
then
	# See job_{rel,ava}.pbs to know which IFUNS and ALGOS must be plotted
	declare -a IFUNS=(
		STR    # RES IFUN: structural
		MCS    # RES IFUN: minimal cut sets
		MCSN   # RES IFUN: normalised MCS
	)
	declare -a ALGOS_REL=(
    	RES_2
    	RES_5
		RES_ES	
    	FE_8
    	FE_16
    	FE_ES
	)
	declare -a ALGOS_AVA=(
    	RES_2
    	RES_5
		RES_ES	
    	RES2_2
    	RES2_5
    	RES2_ES
	)
else
	# See job_{rel,ava}.sh to know which IFUNS and ALGOS will be plotted
	for I in "${!IFUNS[@]}"; do IFUNS[$I]="$I"; done
	for A in "${!ALGOS_REL[@]}"; do ALGOS_REL[$A]="$A"; done
	for A in "${!ALGOS_AVA[@]}"; do ALGOS_AVA[$A]="$A"; done
fi

# Summarise reliability results
for J in "${!JOBS_REL[@]}"; do
	PREFIX="${J}_$RELSUFFIX";
	SUMMARY="${PREFIX}_${SUFFIX}";
	echo >> "$SUMMARY";
	# - first His Lordship crude Monte Carlo
	echo -e "CMC\n" >> "$SUMMARY";
	extract_precision "$PLVL" "${PREFIX}_CMC.out" >> "$SUMMARY"
	echo >> "$SUMMARY";
	# - then each rare event simulation algorithm...
	for A in "${ALGOS_REL[@]}"; do
		echo -e "$A\n" >> "$SUMMARY";
		# ...with each importance function
		for I in "${IFUNS[@]}"; do
			echo -e "IFUN: $I\n" >> "$SUMMARY";
			extract_precision "$PLVL" "${PREFIX}_${A}_$I.out" >> "$SUMMARY";
			echo >> "$SUMMARY";
		done
	done
done

# Summarise availability results
for J in "${!JOBS_AVA[@]}"; do
	PREFIX="${J}_$AVASUFFIX";
	SUMMARY="${PREFIX}_${SUFFIX}";
	echo >> "$SUMMARY";
	# - first His Lordship crude Monte Carlo
	echo -e "CMC\n" >> "$SUMMARY";
	extract_precision "$PLVL" "${PREFIX}_CMC.out" >> "$SUMMARY"
	echo >> "$SUMMARY";
	# - then each rare event simulation algorithm...
	for A in "${ALGOS_AVA[@]}"; do
		echo -e "$A\n" >> "$SUMMARY";
		# ...with each importance function
		for I in "${IFUNS[@]}"; do
			echo -e "IFUN: $I\n" >> "$SUMMARY";
			extract_precision "$PLVL" "${PREFIX}_${A}_$I.out" >> "$SUMMARY";
			echo >> "$SUMMARY";
		done
	done
done

# Prepare CSV file for scatter plots
CSV=${SUFFIX%.txt}.csv;
CSVORDER="CMC";
for A in "${ALGOS_AVA[@]}" "${ALGOS_REL[@]}"; do
	for I in "${IFUNS[@]}"; do
		CSVORDER+=",${A}_$I";
	done
done

# Plot each experiment
source "$DIR_SCRIPTS"/plots.cfg;
echo "Plotting... (see $LOG)";
for f in *"_$SUFFIX"; do
	EXPERIMENT="${f%%_$SUFFIX}";
	if [[ $EXPERIMENT =~ $RELSUFFIX ]]; then
		setup_plot_rel $EXPERIMENT;
	else
		setup_plot_ava $EXPERIMENT;
	fi
	# First create .dat and .csv files from summaries
	python3 "$DIR_SCRIPTS"/make_dat_and_csv.py "$f" "_$SUFFIX" "$CSV" "$CSVORDER" 2>&1 >> "$LOG";
	# Then use the .dat files for GNU-plotting
	echo "Plotting $EXPERIMENT" 2>&1 >> "$LOG";
	EXPERIMENT=$EXPERIMENT gnuplot "$DIR_SCRIPTS"/plot_CI_precision.gp 2>&1 >> "$LOG";
	echo "--" >> "$LOG";
	# If possible create PDF from SVG file
	if [ -x "$(command -v svg2pdf)" ]; then
		svg2pdf "$EXPERIMENT".{svg,pdf} 2>/dev/null;
	fi
done

# Prepare CSV for scatter plots
sed -i "s/\<0\.0\>/$FAIL_VALUE/g" "$CSV";
# If possible create plots by compiling the TEX
if [ -x "$(command -v pdflatex)" ]; then
	ln -s "${DIR_SCRIPTS}"/scatterplots.tex;
	pdflatex scatterplots.tex 2>&1 >/dev/null;
	rm scatterplots.{aux,tex};
fi

# Organise the directory structure
mkdir -p raw data plots summaries;
mv *.{out,err} ./raw/   2>/dev/null;
mv *.{dat,csv} ./data/  2>/dev/null;
mv *.{svg,pdf} ./plots/ 2>/dev/null;
mv *.txt ./summaries/   2>/dev/null;

echo "All plots ready in `pwd`/plots";
cd "$CWD";

return 0;

