#!/usr/bin/env bash
#
# Full experimental run (requires PBS-administered cluster)
#
# Author:  Carlos E. Budde
# Date:    16.10.2021
# License: GPLv3
#
#   NOTE: this script should be sourced from another script
#   Cool site: http://wiki.bash-hackers.org

source "$DIR_SCRIPTS/fig_utils.sh";

export NUM_REPS=10;  # number of times each experiment is repeated

export PBS_QUEUE="common_cpuQ";  # PBS queue name (cluster-specific)
export PBS_JOB_LIMIT=30;  # Queue job limit: qstat -Qf $PBS_QUEUE --> max_queued
export JOBPREFIX="FIG";   # prefix for jobs launched in the PBS queue
declare -a JOB_IDS=();

# XXX The keys of the following arrays must match prefixes of IOSA model files

# Jobs for transient analysis
declare -A JOBS_REL;      # XXX plot_last_results.sh reads this array !!
JOBS_REL[DSPARE_3]=300;   #    5 m
JOBS_REL[DSPARE_4]=1800;  #   30 m
JOBS_REL[DSPARE_5]=10800; #    3 h
JOBS_REL[HECS_2]=20;      #   20 s
JOBS_REL[HECS_3]=300;     #    5 m
JOBS_REL[HECS_4]=1800;    #   30 m
JOBS_REL[HECS_5]=10800;   #    3 h
JOBS_REL[FTPP_4]=30;      #   30 s
JOBS_REL[FTPP_5]=240;     #    4 m
JOBS_REL[FTPP_6]=2400;    #   40 m
JOBS_REL[HVC_4]=90;       #   90 s
JOBS_REL[HVC_5]=300;      #    5 m
JOBS_REL[HVC_6]=1800;     #   30 m
JOBS_REL[HVC_7]=7200;     #    2 h
JOBS_REL[RWC_2]=300;      #    5 m
JOBS_REL[RWC_3]=1800;     #   30 m
JOBS_REL[RWC_4]=7200;     #    2 h

# Jobs for steady-state analysis
declare -A JOBS_AVA;      # XXX plot_last_results.sh reads this array !!
JOBS_AVA[HECS_1]=5;       #    5 s
JOBS_AVA[HECS_2]=20;      #   20 s
JOBS_AVA[HECS_3]=120;     #    2 m
JOBS_AVA[HECS_4]=600;     #   10 m
JOBS_AVA[HECS_5]=3600;    #    1 h
JOBS_AVA[VOT_1]=30;       #   30 s
JOBS_AVA[VOT_2]=300;      #    5 m
JOBS_AVA[VOT_3]=1800;     #   30 m
JOBS_AVA[VOT_4]=10800;    #    3 h
JOBS_AVA[RWC_1]=30;       #   30 s
JOBS_AVA[RWC_2]=300;      #    5 m
JOBS_AVA[RWC_3]=1800;     #   30 m
JOBS_AVA[RWC_4]=7200;     #    2 h
JOBS_AVA[RC_3]=30;        #   30 s
JOBS_AVA[RC_4]=300;       #    5 m
JOBS_AVA[RC_5]=1800;      #   30 m
JOBS_AVA[RC_6]=7200;      #    2 h

# Launch each job in an independent PBS node
echo "Launching PBS jobs to study unreliability (transient properties)";
for J in "${!JOBS_REL[@]}"; do
	export JOB="${J}_$RELSUFFIX";
	export TIME=${JOBS_REL[$J]};  # time that each individual simulation runs
	export EXEC="`execution $TIME $NUM_REPS`";
	export MODEL="$DIR_MODELS/${JOB}.sa";
	is_file "$MODEL";
	JOB_IDS+=("$(qsub -q $PBS_QUEUE -V         \
	               -N "${JOBPREFIX}_${JOB}"    \
	               -o "$DIR_RESULTS/$JOB.out"  \
	               -e "$DIR_RESULTS/$JOB.err"  \
	               "$DIR_SCRIPTS"/job_rel.pbs;)")
	echo "  - Submitted job $JOB: ${JOB_IDS[-1]}";
	wait_for_PBS_queue $PBS_JOB_LIMIT "$JOBPREFIX";  # respect the job limit per user
done
echo "Launching PBS jobs to study unavailability (steady-state properties)";
for J in "${!JOBS_AVA[@]}"; do
	export JOB="${J}_$AVASUFFIX";
	export TIME=${JOBS_AVA[$J]};  # time that each individual simulation runs
	export EXEC="`execution $TIME $NUM_REPS`";
	export MODEL="$DIR_MODELS/${JOB}.sa";
	is_file "$MODEL";
	JOB_IDS+=("$(qsub -q $PBS_QUEUE -V         \
	               -N "${JOBPREFIX}_${JOB}"    \
	               -o "$DIR_RESULTS/$JOB.out"  \
	               -e "$DIR_RESULTS/$JOB.err"  \
	               "$DIR_SCRIPTS"/job_ava.pbs;)")
	echo "  - Submitted job $JOB: ${JOB_IDS[-1]}";
	wait_for_PBS_queue $PBS_JOB_LIMIT "$JOBPREFIX";  # respect the job limit per user
done

# Wait for it
echo "Waiting for all PBS jobs to finish";
IDS=("${JOB_IDS[@]%%\.*}");  # strip suffixes from PBS Job IDs
while
	RUNNING=`qstat ${IDS[@]} 2>/dev/null | wc -l 2>/dev/null`
	[ $RUNNING -gt 0 ]
do
	sleep 30;
done

return 0;

